/*
 * vcvs_acltest.c:
 * Access-control-list check for VSCVS.
 *
 * This is designed to be invoked through the CVSROOT/commitinfo mechanism
 * when a user attempts to commit files.
 *
 * Copyright (c) 2002 Chris Lightfoot. All rights reserved.
 * Email: chris@ex-parrot.com; WWW: http://www.ex-parrot.com/~chris/
 *
 */

static const char rcsid[] = "$Id: vcvs_acltest.c,v 1.2 2002/06/04 16:48:39 chris Exp $";

#include <errno.h>
#include <pwd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

/* path_prefix_match:
 * Is path matched by prefix. */
int path_prefix_match(const char *path, const char *prefix) {
    if (strlen(prefix) > strlen(path))
        return 0;
    else return (!(strncmp(prefix, path, strlen(prefix))));
}

int main(int argc, char *argv[]) {
    char *acltab, *repos, *user, *buf;
    size_t nbuf, max_path_match = 0;
    int reject = 1;
    struct passwd *pw;
    FILE *fp;
    
    /* ACL file is specified as first argument. */
    if (!(acltab = argv[1])) {
        fprintf(stderr, "vcvs_acltest: wrong number of arguments\n");
        return -1;
    }
    
    /* Repository is specified as second argument. */
    if (!(repos = argv[2])) {
        fprintf(stderr, "vcvs_acltest: wrong number of arguments\n");
        return -1;
    }
    
    /* Use the CVS_USER argument supplied by cvs pserver if available;
     * otherwise we are running locally, so use the actual user-ID. */
    if (!(user = getenv("CVS_USER")) && !((pw = getpwuid(getuid())) && (user = pw->pw_name))) {
        fprintf(stderr, "vcvs_acltest: who are you?\n");
        return -1;
    }

    /* Open ACL file. This consists of a list of lines of the form
     *  path user [user ...]
     * 
     * The longest path which matches is used. */
    if (!(fp = fopen(acltab, "rt"))) {
        fprintf(stderr, "vcvs_acltest: %s: %s\n", acltab, strerror(errno));
        return -1;
    }

    nbuf = 1024;
    buf = malloc(nbuf);

    while (fgets(buf, nbuf, fp)) {
        char *p, *q;
        char *path;

        /* Read a full line. */
        if (buf[strlen(buf) - 1] != '\n') {
            do
                buf = realloc(buf, nbuf *= 2);
            while (fgets(buf + strlen(buf), nbuf - strlen(buf), fp)
                   && buf[strlen(buf) - 1] != '\n');
        }

        if (buf[strlen(buf) - 1] == '\n')
            buf[strlen(buf) - 1] = 0;

        /* Discard comment */
        p = strchr(buf, '#');
        if (p) *p = 0;

        /* Skip leading whitespace. */
        p = buf + strspn(buf, " \t");

        /* Find path. */
        q = p + strcspn(p, " \t");

        /* Test path for ACL. */
        if (q > p) {
            size_t pathlen;
            path = p;
            *q = 0;
            if (*(q - 1) == '/')
                *(q - 1) = 0;
            pathlen = strlen(path);
            if (pathlen >= max_path_match && path_prefix_match(repos, path)) {
                /* See if this ACL matches this user. */
                int matched = 0;
                do {
                    p = q + 1;
                    p += strspn(p, " \t");
                    q = p + strcspn(p, " \t");
                    if (q > p) {
                        *q = 0;
                        if (strcmp(p, user) == 0) {
                            matched = 1;
                            break;
                        }
                    }
                } while (*p);

                /* Allow multi-line ACLs. */
                if (!matched && pathlen > max_path_match)
                    reject = 1;
                else if (matched)
                    reject = 0;

                max_path_match = pathlen;
            }
        }
    }

    if (reject)
        fprintf(stderr, "vcvs_acltest: %s: permission denied\n", repos);

    return reject;
}
